from enum import auto

import numpy as np

from ..events.dataclass import Property, evented_dataclass
from ..misc import StringEnum
from .colorbars import make_colorbar
from .standardize_color import transform_color


class ColormapInterpolationMode(StringEnum):
    """INTERPOLATION: Interpolation mode for colormaps.

    Selects an interpolation mode for the colormap.
            * linear: colors are defined by linear interpolation between
              colors of neighboring controls points.
            * zero: colors are defined by the value of the color in the
              bin between by neighboring controls points.
    """

    LINEAR = auto()
    ZERO = auto()


@evented_dataclass
class Colormap:
    """Colormap that relates intensity values to colors.

    Attributes
    ----------
    colors : array, shape (N, 4)
        Data used in the colormap.
    name : str
        Name of the colormap.
    controls : array, shape (N,) or (N+1,)
        Control points of the colormap.
    interpolation : str
        Colormap interpolation mode, either 'linear' or
        'zero'. If 'linear', ncontrols = ncolors (one
        color per control point). If 'zero', ncontrols
        = ncolors+1 (one color per bin).
    """

    colors: Property[np.ndarray, None, transform_color]
    name: str = 'custom'
    controls: Property[np.ndarray, None, np.asarray] = np.zeros((0, 4))
    interpolation: Property[
        ColormapInterpolationMode, str, ColormapInterpolationMode
    ] = ColormapInterpolationMode.LINEAR

    def __post_init__(self):
        if len(self.controls) == 0:
            n_controls = len(self.colors) + int(
                self._interpolation == ColormapInterpolationMode.ZERO
            )
            self.controls = np.linspace(0, 1, n_controls)

    def __iter__(self):
        yield from (self.colors, self.controls, self.interpolation)

    def map(self, values):
        values = np.atleast_1d(values)
        if self._interpolation == ColormapInterpolationMode.LINEAR:
            # One color per control point
            cols = [
                np.interp(values, self.controls, self.colors[:, i])
                for i in range(4)
            ]
            cols = np.stack(cols, axis=1)
        elif self._interpolation == ColormapInterpolationMode.ZERO:
            # One color per bin
            indices = np.clip(
                np.searchsorted(self.controls, values) - 1, 0, len(self.colors)
            )
            cols = self.colors[indices.astype(np.int32)]
        else:
            raise ValueError('Unrecognized Colormap Interpolation Mode')

        return cols

    @property
    def colorbar(self):
        return make_colorbar(self)



import dataclasses as _dc
import operator
import typing
import warnings
from contextlib import contextmanager
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    NamedTuple,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

import dask.array as da
import numpy as np
import toolz as tz
from typing_extensions import (
    Annotated,
    _AnnotatedAlias,
    get_args,
    get_origin,
    get_type_hints,
)

try:
    from typing import _tp_cache
except ImportError:

    def _tp_cache(x):
        return x


from .event import EmitterGroup, Event
from .types import SupportsEvents

ON_SET = "_on_{name}_set"
ON_GET = "_on_{name}_get"
C = TypeVar("C")
T = TypeVar("T")
NO_ATTR = object()


class Property:
    """Declare a dataclass field as a property with getter/setter functions.

    This class is nearly an alias for `typing_extensions.Annotated` (or, in
    ≥python3.9, `typing.Annotated`), except that it provides stricter type
    checking.  See `__class_getitem__` for special considerations.

    ``Property`` should be used with the actual type, followed by an optional
    ``getter`` function (or ``None``), followed by an optional ``setter``
    function (or ``None``): e.g. ``Property[Blending, str, Blending]``

    Examples
    --------

    >>> @evented_dataclass
    ... class Test:
    ...     # x will be stored as an int but *retrieved* as a string.
    ...     x: Property[int, str, int]
    """

    def __new__(cls, *args, **kwargs):
        raise TypeError("Type Property cannot be instantiated")

    def __init_subclass__(cls, *args, **kwargs):
        raise TypeError(f"Cannot subclass {cls.__module__}.Property")

    @_tp_cache
    def __class_getitem__(cls, params):
        """Create the runtime type representation of a Property type.

        Unfortunately, in order for this to behave like an `Annotated` type,
        without us having to vendor a lot of the python
        typing library, this must return an instance of
        `typing_extensions._AnnotatedAlias` (which just points to
        `typing._AnnotatedAlias` after python 3.9).  Hence the private import.
        Since this works on python3.9 (when typing acquired the Annotated type)
        it seems unlikely to break in future python releases.  But it should be
        monitored.
        """
        if not isinstance(params, tuple) or not (1 < len(params) < 4):
            raise TypeError(
                "Property[...] should be used with exactly two or three "
                "arguments (a type, a getter, and an optional setter)"
            )
        msg = "Property[T, ...]: T must be a type."
        origin = typing._type_check(params[0], msg)
        if params[1] is not None and not callable(params[1]):
            raise TypeError(f"Property getter not callable: {params[1]}")
        if (
            len(params) > 2
            and params[2] is not None
            and not callable(params[2])
        ):
            raise TypeError(f"Property getter not callable: {params[1]}")
        metadata = tuple(params[1:])
        return _AnnotatedAlias(origin, metadata)


# from dataclasses.py in the standard library
def _is_classvar(a_type):
    # This test uses a typing internal class, but it's the best way to
    # test if this is a ClassVar.
    return a_type is typing.ClassVar or (
        type(a_type) is typing._GenericAlias
        and a_type.__origin__ is typing.ClassVar
    )


# from dataclasses.py in the standard library
def _is_initvar(a_type):
    # The module we're checking against is the module we're
    # currently in (dataclasses.py).
    return a_type is _dc.InitVar or type(a_type) is _dc.InitVar


@contextmanager
def stripped_annotated_types(cls):
    """Temporarily strip Annotated types (for cleaner function signatures)."""
    original_annotations = cls.__annotations__
    cls.__annotations__ = get_type_hints(cls, include_extras=False)
    yield
    cls.__annotations__ = original_annotations


def set_with_events(self: C, name: str, value: Any) -> None:
    """Modified __setattr__ method that emits an event when set.

    Events will *only* be emitted if the ``name`` of the attribute being set
    is one of the dataclass fields (i.e. ``name in self.__annotations__``),
    and the dataclass ``__post_init__` method has already been called.

    Also looks for and calls an optional ``_on_name_set()`` method afterwards.

    Order of operations:
        1. Call the original ``__setattr__`` function to set the value
        2. Look for an ``_on_name_set`` method on the object
            a. If present, call it with the current value
            b. That method can do anything (including changing the value, or
               emitting its own events if necessary).  If changing the value,
               it should check to make sure that it is different than the
               current value before setting, or a ``RecursionError`` may occur.
            c. If that method returns ``True``. Return *without* emitting
               an event.
        3. If ``_on_name_set`` has not returned ``True``, then emit an event
           from the EventEmitter with the corresponding ``name`` in the.
           e.g. ``self.events.<name>(value=value)``.

    Parameters
    ----------
    self : C
        An instance of the decorated dataclass of Type[C]
    name : str
        The name of the attribute being set.
    value : Any
        The new value for the attribute.
    fields : set of str
        Only emit events for field names in this set.
    """
    if name not in getattr(self, 'events', {}):
        # fallback to default behavior
        object.__setattr__(self, name, value)
        return

    # grab current value
    before = getattr(self, name, NO_ATTR)
    object.__setattr__(self, name, value)

    # if custom set method `_on_<name>_set` exists, call it
    setter_method = getattr(self, ON_SET.format(name=name), None)
    if callable(setter_method):
        # the method can return True, if it wants to handle its own events
        try:
            if setter_method(getattr(self, name)):
                return
        except Exception as e:
            if before is NO_ATTR:
                object.__delattr__(self, name)
            else:
                object.__setattr__(self, name, before)
            meth_name = f"{self.__class__.__name__}.{ON_SET.format(name=name)}"
            raise type(e)(f"Error in {meth_name} (value not set): {e}")

    # if different we emit the event with new value
    after = getattr(self, name)
    if not self.__equality_checks__.get(name, is_equal)(after, before):
        # use gettattr again in case `_on_name_set` has modified it
        getattr(self.events, name)(value=after)  # type: ignore


def is_equal(v1, v2):
    """
    Function for basic comparison compare.

    Parameters
    ----------
    v1: Any
        first value
    v2: Any
        second value

    Returns
    -------
    result : bool
        Returns ``True`` if ``v1`` and ``v2`` are equivalent.
        If an exception is raised during comparison, returns ``False``.

    Warns
    -----
    UserWarning
        If an exception is raised during comparison.
    """
    try:
        return bool(v1 == v2)
    except Exception as e:
        warnings.warn(
            "Comparison method failed. Returned False. "
            f"There may be need to define custom compare methods in __equality_checks__ dictionary. Exception {e}"
        )
        return False


def _type_to_compare(type_) -> Optional[Callable[[Any, Any], bool]]:
    """
    Try to determine compare function for types which cannot be compared with `operator.eq`.
    Currently support `numpy.ndarray` and `dask.core.Array`.
    Unpack `typing.Optional` and `dataclasses.InitVar` box.

    Parameters
    ----------
    type_: type
    type to examine

    Returns
    -------
    Optional[Callable[[Any, Any], bool]]
        Compare function
    """
    import inspect

    if isinstance(type_, _AnnotatedAlias):
        type_ = type_.__origin__
    if isinstance(type_, _dc.InitVar):
        type_ = type_.type

    # get main type from Optional[type]
    _args = get_args(type_)
    if (
        get_origin(type_) is Union
        and len(_args) == 2
        and isinstance(None, _args[1])
    ):
        type_ = _args[0]
    if not inspect.isclass(type_):
        if not getattr(type_, "__module__", None) == "typing":
            warnings.warn(f"Bug in type recognition {type_}")
        return
    if issubclass(type_, np.ndarray):
        return np.array_equal
    if issubclass(type_, da.core.Array):
        return operator.is_
    return None


def add_events_to_class(cls: Type[C]) -> Type[C]:
    """Return a new __post_init__ method wrapper with events.

    Parameters
    ----------
    cls : type
        The class being decorated as a dataclass

    Returns
    -------
    Callable[..., None]
        A modified __post_init__ method that wraps the original.
    """

    # get a handle to the original __post_init__ method if present
    orig_post_init: Callable[..., None] = getattr(cls, '__post_init__', None)

    _fields = [
        _dc._get_field(cls, name, type_)
        for name, type_ in cls.__dict__.get('__annotations__', {}).items()
    ]
    e_fields = {
        fld.name: None
        for fld in _fields
        if fld._field_type is _dc._FIELD and fld.metadata.get("events", True)
    }

    # create dict with compare functions for fields which cannot be compared
    # using standard equal operator, like numpy arrays.
    # it will be set to __equality_checks__ class parameter.
    compare_dict_base = getattr(cls, "__equality_checks__", {})
    compare_dict = {
        n: t
        for n, t in {
            name: _type_to_compare(type_)
            for name, type_ in cls.__dict__.get('__annotations__', {}).items()
            if name not in compare_dict_base
        }.items()
        if t is not None  # walrus operator is supported from python 3.8
    }
    # use compare functions provided by class creator.
    compare_dict.update(compare_dict_base)

    def evented_post_init(self: T, *initvars) -> None:
        # create an EmitterGroup with an EventEmitter for each field
        # in the dataclass, skip those with metadata={'events' = False}
        if hasattr(self, 'events') and isinstance(self.events, EmitterGroup):
            for em in self.events.emitters:
                e_fields.pop(em, None)
            self.events.add(**e_fields)
        else:
            self.events = EmitterGroup(
                source=self,
                auto_connect=False,
                **e_fields,
            )
        # call original __post_init__
        if orig_post_init is not None:
            orig_post_init(self, *initvars)

    # modify __setattr__ with version that emits an event when setting
    setattr(cls, '__setattr__', set_with_events)
    setattr(cls, '__post_init__', evented_post_init)
    setattr(cls, '__equality_checks__', compare_dict)
    return cls


# make the actual getter/setter functions that the property will use
@tz.curry
def prop_getter(priv_name: str, fcoerce: Callable, obj) -> Any:
    # val = object.__getattribute__(obj, name)
    value = fcoerce(getattr(obj, priv_name))
    pub_name = priv_name.lstrip("_")
    getter_method = getattr(obj, ON_GET.format(name=pub_name), None)
    if callable(getter_method):
        return getter_method(value)
    return value


@tz.curry
def prop_setter(
    priv_name: str, default: Any, fcoerce: Callable, obj, value
) -> None:
    # during __init__, the dataclass will try to set the instance
    # attribute to the property itself!  So we intervene and set it
    # to the default value from the dataclass declaration
    pub_name = priv_name.lstrip("_")
    if type(value) is property:
        value = default
        # dataclasses may define attributes as field(...)
        # so we need to get the default value from the field
        if isinstance(default, _dc.Field):
            default = cast(_dc.Field, default)
            # there will only ever be default_factory OR default
            # otherwise an exception will have been raised earlier.
            if default.default_factory is not _dc.MISSING:
                value = default.default_factory()
            elif default.default is not _dc.MISSING:
                value = default.default
        # If the value is still missing, then it means that the user
        # failed to provide a required positional argument when
        # instantiating the dataclass.
        if value is _dc.MISSING:
            raise TypeError(
                "__init__() missing required "
                f"positional argument: '{pub_name}'"
            )
    setattr(obj, priv_name, fcoerce(value))


class TypeGetSet(NamedTuple):
    '''A simple named tuple with a type, getter func, and setter func.'''

    type: Type[T]
    fget: Callable[[T], Any]
    fset: Callable[[Any], T]


@tz.curry
def _try_coerce(func, name, value):
    if func is not None:
        try:
            return func(value)
        except Exception as e:
            raise TypeError(f"Failed to coerce value {value} in {name}: {e}")
    return value


def parse_annotated_types(
    cls: Type,
) -> Dict[str, TypeGetSet]:
    """Return a dict of field names and their types.

    If any type annotations are instances of ``Property``, then their
    paramereters will be used to create type-coercion functions that will be
    called during getting/setting of this field.

    Parameters
    ----------
    cls : Type
        The class being processed as a dataclass

    Returns
    -------
    dict
        A dict of field names to ``TypeGetSet`` objects
    """
    out: Dict[str, TypeGetSet] = {}
    for name, typ in get_type_hints(cls, include_extras=True).items():
        d = [typ, None, None]
        if get_origin(typ) is Annotated:
            args = get_args(typ)
            d[: len(args)] = args
        d[1] = _try_coerce(d[1], name)
        d[2] = _try_coerce(d[2], name)
        out[name] = TypeGetSet(*d)
    return out


def convert_fields_to_properties(cls: Type[C]) -> Type[C]:
    """Convert all fields in a dataclass instance to property descriptors.

    Note: this modifies class Type[C] (the class that was decorated with
    ``@dataclass``) *after* instantiation of the class.  In other words, for a
    given field `f` on class `C`, `C.f` will *not* be a property descriptor
    until *after* C has been instantiated: `c = C()`.  (And reminder: property
    descriptors are class attributes).

    The reason for this is that dataclasses can have "default factory"
    functions that create default values for fields only during instantiation,
    and we don't want to have to recreate that logic here, (but we do need to
    know what the value of the field is).

    Parameters
    ----------
    obj : T
        An instance of class ``Type[C]`` that has been deorated as a dataclass.
    """
    from numpydoc.docscrape import ClassDoc

    # grab docstring to populate properties docs
    cls_doc = ClassDoc(cls)
    params = {p.name: p for p in cls_doc["Parameters"]}

    coerce_funcs = parse_annotated_types(cls)
    # loop through annotated members of the glass
    for name, type_ in list(cls.__dict__.get('__annotations__', {}).items()):
        # ClassVar and InitVar types are exempt from dataclasses and properties
        # https://docs.python.org/3/library/dataclasses.html#class-variables
        if _is_classvar(type_) or _is_initvar(type_):
            continue
        private_name = f"_{name}"
        # store the original value for the property
        default = getattr(cls, name, _dc.MISSING)

        # add the private_name as a ClassVar annotation on the original class
        # (annotations of type ClassVar are ignored by the dataclass)
        # `self.x` is the property, `self._x` contains the data
        cls.__annotations__[private_name] = ClassVar[type_]

        if name in coerce_funcs:
            fget = prop_getter(private_name, coerce_funcs[name].fget)
            fset = prop_setter(private_name, default, coerce_funcs[name].fset)
        else:
            fget = _try_coerce(None, name)
            fset = _try_coerce(None, name)
        # bring the docstring from the class to the property
        doc = None
        if name in params:
            param = params[name]
            doc = "\n".join(param.desc)
            # TODO: could compare param.type to field.type here for consistency
            # alternatively, we may just want to use pydantic for type
            # validation.

        # create the actual property descriptor
        prop = property(fget=fget, fset=fset, fdel=None, doc=doc)
        setattr(cls, name, prop)
    return cls


def update_from_dict(self, values, compare_fun=is_equal):
    if isinstance(values, self.__class__):
        values = values.asdict()
    if not isinstance(values, dict):
        raise ValueError(f"Unsupported update from {type(values)}")
    if all(compare_fun(values[k], v) for k, v in self.asdict().items()):
        return

    if isinstance(self, SupportsEvents):
        with self.events.blocker():
            for key, value in values.items():
                setattr(self, key, value)
        self.events(Event(self))
    else:
        for key, value in values.items():
            setattr(self, key, value)


@tz.curry
def evented_dataclass(
    cls: Type[C],
    *,
    init: bool = True,
    repr: bool = True,
    eq: bool = True,
    order: bool = False,
    unsafe_hash: bool = False,
    frozen: bool = False,
    events: bool = True,
    properties: bool = True,
) -> Type[C]:
    """Enhanced dataclass decorator with events and property descriptors.

    Examines PEP 526 __annotations__ to determine fields.  Fields are defined
    as class attributes with a type annotation.  Everything but ``events`` and
    ``properties`` are defined on the builtin dataclass decorator.

    Note: if ``events==False`` and ``properties==False``, this is functionally
    equivalent to the builtin ``dataclasses.dataclass``

    Parameters
    ----------
    cls : Type[C]
        [description]
    init : bool, optional
        If  true, an __init__() method is added to the class, by default True
    repr : bool, optional
        If true, a __repr__() method is added, by default True
    eq : bool, optional
        [description], by default True
    order : bool, optional
        If true, rich comparison dunder methods are added, by default False
    unsafe_hash : bool, optional
        If true, a __hash__() method function is added, by default False
    frozen : bool, optional
        If true, fields may not be assigned to after instance creation, by
        default False
    events : bool, optional
        If true, an ``EmmitterGroup`` instance is added as attribute "events".
        Events will be emitted each time one of the dataclass fields are
        altered, by default True
    properties : bool, optional
        If true, field attributes will be converted to property descriptors.
        If the class has a class docstring in numpydocs format, docs for each
        property will be taken from the ``Parameters`` section for the
        corresponding parameter, by default True

    Returns
    -------
    decorated class
        Returns the same class as was passed in, with dunder methods
        added based on the fields defined in the class.

    Raises
    ------
    ValueError
        If both ``properties`` and ``frozen`` are True
    """

    if frozen and (events or properties):
        raise ValueError(
            "`frozen=True` is incompatible `properties=True` or `events=True`"
        )

    # TODO: currently, events must be process first here, otherwise
    # metada={'events':False} does not work... but that should be fixed.
    if events:
        # create a modified __post_init__ method that creates an EmitterGroup
        cls = add_events_to_class(cls)

    # if neither events or properties are True, this function is exactly like
    # the builtin `dataclasses.dataclass`
    with stripped_annotated_types(cls):
        cls = _dc.dataclass(  # type: ignore
            cls,
            init=init,
            repr=repr,
            eq=eq,
            order=order,
            unsafe_hash=unsafe_hash,
            frozen=frozen,
        )

    if properties:
        # convert public dataclass fields to properties
        cls = convert_fields_to_properties(cls)
    setattr(cls, 'asdict', _dc.asdict)
    setattr(cls, 'update', update_from_dict)
    return cls



import os
import sys
from inspect import isclass, signature
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    List,
    Sequence,
    Tuple,
    Type,
    Union,
)
from warnings import warn

from magicgui import magicgui
from napari_plugin_engine import HookImplementation, PluginManager

from ..types import AugmentedFunction, AugmentedWidget
from ..utils._appdirs import user_site_packages
from ..utils.misc import camel_to_spaces, running_as_bundled_app
from . import _builtins, hook_specifications

if sys.platform.startswith('linux') and running_as_bundled_app():
    sys.path.append(user_site_packages())


if TYPE_CHECKING:
    from qtpy.QtWidgets import QWidget


# the main plugin manager instance for the `napari` plugin namespace.
plugin_manager = PluginManager(
    'napari', discover_entry_point='napari.plugin', discover_prefix='napari_'
)
with plugin_manager.discovery_blocked():
    plugin_manager.add_hookspecs(hook_specifications)
    plugin_manager.register(_builtins, name='builtins')


dock_widgets: Dict[Tuple[str, str], Tuple[Type['QWidget'], dict]] = dict()
function_widgets: Dict[Tuple[str, str], Tuple[Callable, dict, dict]] = dict()


def register_dock_widget(
    args: Union[AugmentedWidget, List[AugmentedWidget]],
    hookimpl: HookImplementation,
):
    from qtpy.QtWidgets import QWidget

    plugin_name = hookimpl.plugin_name
    hook_name = '`napari_experimental_provide_dock_widget`'
    for arg in args if isinstance(args, list) else [args]:
        if isinstance(arg, tuple):
            if not arg:
                warn(
                    f'Plugin {plugin_name!r} provided an invalid tuple to '
                    f'{hook_name}.  Skipping'
                )
                continue
            _cls = arg[0]
            kwargs = arg[1] if len(arg) > 1 else {}
        else:
            _cls, kwargs = (arg, {})

        if not (isclass(_cls) and issubclass(_cls, QWidget)):
            warn(
                f'Plugin {plugin_name!r} provided an invalid '
                f'widget type to {hook_name}: {_cls!r}. Widget ignored.'
            )
            continue

        if not isinstance(kwargs, dict):
            warn(
                f'Plugin {plugin_name!r} provided invalid kwargs '
                f'to {hook_name} for class {_cls.__name__}. Widget ignored.'
            )
            continue

        # Get widget name
        name = str(kwargs.get('name', '')) or camel_to_spaces(_cls.__name__)

        key = (plugin_name, name)
        if key in dock_widgets:
            warn(
                "Plugin '{}' has already registered a dock widget '{}' "
                'which has now been overwritten'.format(*key)
            )
        dock_widgets[key] = (_cls, kwargs)


magicgui_sig = {
    name
    for name, p in signature(magicgui).parameters.items()
    if p.kind is p.KEYWORD_ONLY
}


def register_function_widget(
    args: Union[AugmentedFunction, List[AugmentedFunction]],
    hookimpl: HookImplementation,
):
    plugin_name = hookimpl.plugin_name
    hook_name = '`napari_experimental_provide_function_widget`'
    for arg in args if isinstance(args, list) else [args]:
        if isinstance(arg, tuple):
            if not arg:
                warn(
                    f'Plugin {plugin_name!r} provided an invalid tuple to '
                    f'{hook_name}. Skipping'
                )
                continue
            func = arg[0]
            magic_kwargs = arg[1] if len(arg) > 1 else {}
            dock_kwargs = arg[2] if len(arg) > 2 else {}  # type: ignore
        else:
            func, magic_kwargs, dock_kwargs = (arg, {}, {})

        if not callable(func):
            msg = (
                f'Plugin {plugin_name!r} provided a non-callable type to '
                f'{hook_name}: {type(func)!r}. Function widget ignored.'
            )
            if isinstance(func, tuple):
                msg += (
                    " To provide multiple function widgets please use "
                    "a LIST of tuples"
                )
            warn(msg)
            continue

        if not isinstance(magic_kwargs, dict):
            warn(
                f'Plugin {plugin_name!r} provided invalid type '
                f'({type(magic_kwargs)}) for 2nd argument to {hook_name} '
                f'for function {func.__name__!r}. Widget ignored.'
            )
            continue

        valid_magic_kwargs = set(signature(func).parameters) | magicgui_sig
        extra_kwargs = set(magic_kwargs) - valid_magic_kwargs
        if extra_kwargs:
            warn(
                f'Plugin {plugin_name!r} provided invalid magicgui kwargs '
                f'to {hook_name} for function {func.__name__!r}: '
                f'{extra_kwargs}. Valid kwargs are {valid_magic_kwargs}.'
                'Widget ignored.'
            )
            continue

        if not isinstance(dock_kwargs, dict):
            warn(
                f'Plugin {plugin_name!r} provided invalid type '
                f'({type(dock_kwargs)}) for 3nd argument to {hook_name} '
                f'for function {func.__name__!r}. Widget ignored.'
            )
            continue

        # Get function name
        name = str(dock_kwargs.get('name', '')) or func.__name__.replace(
            '_', ' '
        )

        key = (plugin_name, name)
        if key in function_widgets:
            warn(
                "Plugin '{}' has already registered a function widget '{}' "
                'which has now been overwritten'.format(*key)
            )

        function_widgets[key] = (func, magic_kwargs, dock_kwargs)


plugin_manager.hook.napari_experimental_provide_dock_widget.call_historic(
    result_callback=register_dock_widget, with_impl=True
)


plugin_manager.hook.napari_experimental_provide_function_widget.call_historic(
    result_callback=register_function_widget, with_impl=True
)

#: Template to use for namespacing a plugin item in the menu bar
menu_item_template = '{}: {}'

__all__ = ["PluginManager", "plugin_manager", 'menu_item_template']



import os
import sys
from contextlib import contextmanager

from qtpy.QtCore import Qt
from qtpy.QtGui import QIcon, QPixmap
from qtpy.QtWidgets import QApplication, QSplashScreen

from napari import __version__

from ..utils.perf import perf_config
from .exceptions import ExceptionHandler
from .qthreading import wait_for_workers_to_quit

NAPARI_ICON_PATH = os.path.join(
    os.path.dirname(__file__), '..', 'resources', 'logo.png'
)
NAPARI_APP_ID = f'napari.napari.viewer.{__version__}'


def set_app_id(app_id):
    if os.name == "nt" and app_id and not getattr(sys, 'frozen', False):
        import ctypes

        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(app_id)


_defaults = {
    'app_name': 'napari',
    'app_version': __version__,
    'icon': NAPARI_ICON_PATH,
    'org_name': 'napari',
    'org_domain': 'napari.org',
    'app_id': NAPARI_APP_ID,
}


def get_app(
    *,
    app_name: str = None,
    app_version: str = None,
    icon: str = None,
    org_name: str = None,
    org_domain: str = None,
    app_id: str = None,
) -> QApplication:
    """Get or create the Qt QApplication.

    There is only one global QApplication instance, which can be retrieved by
    calling get_app again, (or by using QApplication.instance())

    Parameters
    ----------
    app_name : str, optional
        Set app name (if creating for the first time), by default 'napari'
    app_version : str, optional
        Set app version (if creating for the first time), by default __version__
    icon : str, optional
        Set app icon (if creating for the first time), by default
        NAPARI_ICON_PATH
    org_name : str, optional
        Set organization name (if creating for the first time), by default
        'napari'
    org_domain : str, optional
        Set organization domain (if creating for the first time), by default
        'napari.org'
    app_id : str, optional
        Set organization domain (if creating for the first time).  Will be
        passed to set_app_id (which may also be called independently), by
        default NAPARI_APP_ID

    Returns
    -------
    QApplication
        [description]

    Notes
    -----
    Substitutes QApplicationWithTracing when the NAPARI_PERFMON env variable
    is set.

    If the QApplication already exists, we call convert_app_for_tracing() which
    deletes the QApplication and creates a new one. However here with get_app
    we need to create the correct QApplication up front, or we will crash
    because we'd be deleting the QApplication after we created QWidgets with
    it, such as we do for the splash screen.
    """
    # napari defaults are all-or nothing.  If any of the keywords are used
    # then they are all used.
    set_values = {k for k, v in locals().items() if v}
    kwargs = locals() if set_values else _defaults

    app = QApplication.instance()
    if app:
        if set_values:
            from warnings import warn

            warn(
                "QApplication already existed, these arguments to to 'get_app'"
                " were ignored: {}".format(set_values)
            )
        if perf_config and perf_config.trace_qt_events:
            from .perf.qt_event_tracing import convert_app_for_tracing

            # no-op if app is already a QApplicationWithTracing
            app = convert_app_for_tracing(app)
        app._existed = True
    else:
        # automatically determine monitor DPI.
        # Note: this MUST be set before the QApplication is instantiated
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)

        if perf_config and perf_config.trace_qt_events:
            from .perf.qt_event_tracing import QApplicationWithTracing

            app = QApplicationWithTracing(sys.argv)
        else:
            app = QApplication(sys.argv)

        # if this is the first time the Qt app is being instantiated, we set
        # the name, so that we know whether to raise_ in Window.show()

        app.setApplicationName(kwargs.get('app_name'))
        app.setApplicationVersion(kwargs.get('app_version'))
        app.setOrganizationName(kwargs.get('org_name'))
        app.setOrganizationDomain(kwargs.get('org_domain'))
        app.setWindowIcon(QIcon(kwargs.get('icon')))
        set_app_id(kwargs.get('app_id'))

    if perf_config and not perf_config.patched:
        # Will patch based on config file.
        perf_config.patch_callables()

    # see docstring of `wait_for_workers_to_quit` for caveats on killing
    # workers at shutdown.
    app.aboutToQuit.connect(wait_for_workers_to_quit)

    return app


@contextmanager
def gui_qt(*, startup_logo=False, gui_exceptions=False, force=False):
    """Start a Qt event loop in which to run the application.

    Parameters
    ----------
    startup_logo : bool, optional
        Show a splash screen with the napari logo during startup.
    gui_exceptions : bool, optional
        Whether to show uncaught exceptions in the GUI, by default they will be
        shown in the console that launched the event loop.
    force : bool, optional
        Force the application event_loop to start, even if there are no top
        level widgets to show.

    Notes
    -----
    This context manager is not needed if running napari within an interactive
    IPython session. In this case, use the ``%gui qt`` magic command, or start
    IPython with the Qt GUI event loop enabled by default by using
    ``ipython --gui=qt``.
    """
    splash_widget = None

    app = get_app()
    if startup_logo and app.applicationName() == 'napari':
        pm = QPixmap(NAPARI_ICON_PATH).scaled(
            360, 360, Qt.KeepAspectRatio, Qt.SmoothTransformation
        )
        splash_widget = QSplashScreen(pm)
        splash_widget.show()
        app._splash_widget = splash_widget

    # instantiate the exception handler
    exception_handler = ExceptionHandler(gui_exceptions=gui_exceptions)
    sys.excepthook = exception_handler.handle

    try:
        yield app
    except Exception:
        exception_handler.handle(*sys.exc_info())

    # if the application already existed before this function was called,
    # there's no need to start it again.  By avoiding unnecessary calls to
    # ``app.exec_``, we avoid blocking.
    # we add 'magicgui' so that anyone using @magicgui *before* calling gui_qt
    # will also have the application executed. (a bandaid for now?...)
    # see https://github.com/napari/napari/pull/2016
    if app.applicationName() in ('napari', 'magicgui'):
        if splash_widget and startup_logo:
            splash_widget.close()
        run(force=force, _func_name='gui_qt')


def run(*, force=False, _func_name='run'):
    """Start the Qt Event Loop

    Parameters
    ----------
    force : bool, optional
        Force the application event_loop to start, even if there are no top
        level widgets to show.
    _func_name : str, optional
        name of calling function, by default 'run'.  This is only here to
        provide functions like `gui_qt` a way to inject their name into the
        warning message.

    Raises
    ------
    RuntimeError
        (To avoid confusion) if no widgets would be shown upon starting the
        event loop.
    """
    app = QApplication.instance()
    if not app:
        raise RuntimeError(
            'No Qt app has been created. '
            'One can be created by calling `get_app()` '
            'or qtpy.QtWidgets.QApplication([])'
        )
    if not app.topLevelWidgets() and not force:
        from warnings import warn

        warn(
            "Refusing to run a QApplication with no topLevelWidgets. "
            f"To run the app anyway, use `{_func_name}(force=True)`"
        )
        return
    app.exec_()
